#!/usr/bin/env python3

import os
import random
import time

from lxml import html
import requests

import click
import toml

# from kaomojitool.kaomoji import KaomojiDBKaomojiExists
# from kaomojitool.kaomoji import KaomojiDBKaomojiDoesntExist
from .kaomoji import Kaomoji
from .kaomoji import KaomojiDB

from .kaomoji import KaomojiDBKaomojiExists
from .kaomoji import KaomojiDBKaomojiDoesntExist

class KaomojiToolNoDatabase(Exception):
    description="Kaomoji edit tool couldn't open database"
    def __init__(self, *args, **kwargs):
        super().__init__(self.description, *args, **kwargs)

DEFAULT_CONFIG = {
    'database_filename': './emoticons.tsv',
}

USER_CONFIG_FILENAME = os.path.expanduser("~/.kaomojitool")
USER_CONFIG: dict

# CONFIG = DEFAULT_CONFIG  # initialize it with defaults

# USER_CONFIG = dict()

# there is:
# 1. DEFAULT_CONFIG: dict
# 2. USER_CONFIG_FILENAME: str
# 3. user_config: dict
# 4. cli_config_filename
# 5. cli_config
# 6. cli_database_filename

class KaomojiTool:

    def __init__(self, cli_database_filename, cli_config_filename, *args, **kwargs):


        self.config = self._update_config(cli_config_filename=cli_config_filename,
                                          cli_database_filename=cli_database_filename)

        self.database = self._open_database()


    def _open_database(self):

        print(self.config)
        database_filename = self.config['database_filename']
        # HERE: also check if it is valid
        if os.path.isfile(database_filename):
            return KaomojiDB(filename=database_filename)
        else:
            raise KaomojiToolNoDatabase

        return None

    def backup_database(self):

        database = self.database
        timestamp = time.time()
        backup_filename = "{filename}.{timestamp}.bkp".\
            format(filename=database.filename, timestamp=timestamp)
        backup = KaomojiDB(filename=database.filename)
        backup.write(filename=backup_filename)

    def _update_config(self, cli_database_filename,
                       cli_config_filename, config=DEFAULT_CONFIG,
                       user_config_filename=USER_CONFIG_FILENAME):
        """Updates config with priority:
        1. user command-line entry.
        2. user default config file.
        3. default config.
        """

        self.config = DEFAULT_CONFIG
        self.user_config = dict()

        if os.path.isfile(USER_CONFIG_FILENAME):
            self.user_config =\
                self._read_config_file(config_filename=USER_CONFIG_FILENAME)
            self.config.update(self.user_config)

        if cli_config_filename and os.path.isfile(cli_config_filename):
            self.user_config =\
                self._read_config_file(config_filename=cli_config_filename)
            self.config.update(self.user_config)

        if cli_database_filename:
            self.config.update({'cli_database_filename':
                                 cli_database_filename})

        return self.config

    def _read_config_file(self, config_filename):

        if os.path.isfile(config_filename):
            self.user_config = toml.load(config_filename)

        return self.user_config

keywords_add_option = click.option(
    "-a", "--add", "keywords_add",
    default=None,
    multiple=True,
    #prompt="Keywords to add, comma-separated",  # this should go inside function
    type=str,
    help = "Comma-separated or option-separated list of keywords to add.")

keywords_remove_option = click.option(
    "-x", "--rm", "keywords_remove",
    default=None,
    multiple=True,
    #prompt="Keywords to remove, comma-separated",  # this should go inside function
    type=str,
    help="Comma-separated or option-separated list of keywords to remove.")

database_filename_option = click.option(
    "-f", "--database", "database_filename",
    default=None,
    type=str,
    help="Kaomoji database file name.")

kaomoji_code_option = click.option(
    "-k", "--kaomoji", "kaomoji_code",
    default=None,
    #prompt="Kaomoji",
    type=str,
    help="Kaomoji; use - to read from STDIN.")

keywords_option = click.option(
    "-w", "--keywords", "keywords",
    default=None,
    #prompt="Keywords, comma-separated",
    type=str,
    help="Comma-separated list of keywords to change.")

config_filename_option = click.option(
    "-c", "--config", "config_filename",
    default=USER_CONFIG_FILENAME,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Kaomoji database file name.")

other_database_filename_option = click.option(
    "-o", "--other-database", "other_database_filename",
    default=None,
    type=str,
    help="Kaomoji database file name to compare with the default.")

diff_type_option = click.option(
    "-t", "--diff-type", "diff_type",
    default="additional",
    #prompt="Keywords, comma-separated",
    type=click.Choice(["additional", "difference", "exclusive",
                       "intersection"], case_sensitive=False),
    help="Comma-separated list of keywords to change.")

query_string_option = click.option(
    "-q", "--query", "query_string",
    default="",
    type=str,
    help="String to query keywords.")

url_to_scrape_option = click.option(
    "-u", "--url", "url_to_scrape",
    default=None,
    type=str,
    help="URL to scrape from"
)
kaomoji_xpath_string_option = click.option(
    "-x", "--kaomoji-xpath-string", "kaomoji_xpath_string",
    default = None,
    type = str,
    help = "XPath of the kaomoji element"
)
keywords_xpath_string_option = click.option(
    "-X", "--keywords-xpath-string", "keywords_xpath_string",
    default=None,
    type=str,
    help="XPath of the keyword element"
)
container_xpath_string_option = click.option(
    "-C", "--container-xpath-string", "container_xpath_string",
    default=None,
    type=str,
    help="XPath of the container element, which contains the kaomoji and it's"\
         " respective kayword(s)"
)

@click.group()
def cli():
    """Toolchain to edit kaomoji database files.

    Contribute at:

    https://github.com/iacchus/kaomoji-database-edit-tool/
    """
    pass

###############################################################################
# add                                                                         #
###############################################################################
@cli.command()
@database_filename_option
@kaomoji_code_option
@keywords_option
@config_filename_option
def add(database_filename, kaomoji_code, keywords, config_filename):
    """Adds the selected kaomoji to the selected database"""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    print("Backing up the database '{}'...".\
          format(kaomojitool.database.filename))
    kaomojitool.backup_database()

    kaomoji_add_list: list = list()

    if kaomoji_code:
        new_kaomoji = Kaomoji(code=kaomoji_code, keywords=keywords)
        kaomoji_add_list.append(new_kaomoji)
    else:  # read from stdin
        stdin = click.get_text_stream('stdin')

        for line in stdin.readlines():
            new_kaomoji = Kaomoji(line_entry=line)
            kaomoji_add_list.append(new_kaomoji)

    for kaomoji in kaomoji_add_list:

        db_kaomoji = kaomojitool.database.get_kaomoji(kaomoji)  # reference or None

        if db_kaomoji and kaomoji.keywords:
            print("Kaomoji already exists! Updating keywords..")
            db_kaomoji.add_keywords(keywords)
        elif not db_kaomoji:
            print("New kaomoji! Adding...")
            db_kaomoji = kaomojitool.database.add_kaomoji(kaomoji)  # returns reference

        print("kaomoji:", db_kaomoji.code)
        print("keywords:", db_kaomoji.keywords)

    print("Writing db", kaomojitool.database.filename)
    kaomojitool.database.write()


###############################################################################
# edit                                                                        #
###############################################################################
@cli.command()
@database_filename_option
@kaomoji_code_option
@keywords_add_option
@keywords_remove_option
@config_filename_option
def edit(database_filename, kaomoji_code, keywords_add, keywords_remove,
         config_filename):
    """Edits the selected kaomoji to the selected database, adding or removing
    keywords.
    """

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    keywords_to_add = ",".join(keywords_add)  # adding will have preemptiness
    keywords_to_remove = ",".join(keywords_remove)

    print("Keywords to remove:", keywords_to_remove)

    edit_kaomoji = kaomojitool.database.get_kaomoji(by_entity=kaomoji_code)

    if not edit_kaomoji:
        print("New kaomoji! Adding it do database...")
        kaomoji_to_add = Kaomoji(code=kaomoji_code)
        edit_kaomoji = kaomojitool.database.add_kaomoji(kaomoji_to_add)
    else:
        print("Kaomoji already exists! Editing keywords int it...")

    if keywords_to_remove:
        edit_kaomoji.remove_keywords(keywords=keywords_to_remove)
    if keywords_to_add:
        edit_kaomoji.add_keywords(keywords=keywords_to_add)

    print("Backing up the database...")
    kaomojitool.backup_database(db=kaomojitool.database)

    print("Editing...")
    print("kaomoji:", edit_kaomoji.code)
    print("keywords:", edit_kaomoji.keywords)
    kaomojitool.database.update_kaomoji(edit_kaomoji)

    print("Writing db", kaomojitool.database.filename)
    kaomojitool.database.write()


###############################################################################
# rm                                                                          #
###############################################################################
@cli.command()
@database_filename_option
@kaomoji_code_option
@config_filename_option
def rm(database_filename, kaomoji_code, config_filename):
    """Removes the selected kaomoji from the selected database"""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    if kaomoji_code:
        kaomoji_to_remove = Kaomoji(code=kaomoji_code)

    if kaomojitool.database.kaomoji_exists(kaomoji_to_remove):
        print("Backing up the database...")
        kaomojitool.backup_database(db=kaomojitool.database)

        print("Removing...")
        print("kaomoji:", kaomoji_to_remove.code)
        print("keywords:", kaomoji_to_remove.keywords)
        kaomojitool.database.remove_kaomoji(kaomoji_to_remove)
    else:
        raise kaomojitool.databaseKaomojiDoesntExist

    print("Writing db", kaomojitool.database.filename)
    kaomojitool.database.write()


###############################################################################
# kwadd                                                                       #
###############################################################################
@cli.command()
@database_filename_option
@kaomoji_code_option
@config_filename_option
@keywords_option
def kwadd(database_filename, kaomoji_code, keywords, config_filename):
    """Add keywords to the selected kaomoji."""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    if not kaomojitool.database.get_kaomoji(by_entity=kaomoji_code):
        print("New kaomoji! Adding it do database...")
        new_kaomoji = Kaomoji(code=kaomoji_code, keywords=keywords)
        kaomojitool.database.add_kaomoji(kaomoji=new_kaomoji)
    else:
        print("Kaomoji already exists! Removing keywords from it...")

    #edit_kaomoji = kaomojitool.database.get_kaomoji_by_code(code=kaomoji_code)
    edit_kaomoji = kaomojitool.database.get_kaomoji(by_entity=kaomoji_code)
    edit_kaomoji.add_keywords(keywords=keywords)

    print("Backing up the database...")
    kaomojitool.backup_database(db=kaomojitool.database)

    print("Removing keywords...")
    print("kaomoji:", edit_kaomoji.code)
    print("keywords:", edit_kaomoji.keywords)
    kaomojitool.database.update_kaomoji(edit_kaomoji)

    print("Writing db", kaomojitool.database.filename)
    kaomojitool.database.write()


###############################################################################
# kwrm                                                                        #
###############################################################################
@cli.command()
@database_filename_option
@kaomoji_code_option
@config_filename_option
@keywords_option
def kwrm(database_filename, kaomoji_code, keywords, config_filename):
    """Remove keywords to the selected kaomoji."""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    if not kaomojitool.database.get_kaomoji(by_entity=kaomoji_code):
        print("New kaomoji! Adding it do database...")
        new_kaomoji = Kaomoji(code=kaomoji_code, keywords=keywords)
        kaomojitool.database.add_kaomoji(kaomoji=new_kaomoji)
    else:
        print("Kaomoji already exists! Adding keywords to it...")

    edit_kaomoji = kaomojitool.database.get_kaomoji(by_entity=kaomoji_code)
    edit_kaomoji.remove_keywords(keywords=keywords)

    print("Backing up the database...")
    kaomojitool.backup_database()

    print("Removing keywords...")
    print("kaomoji:", edit_kaomoji.code)
    print("keywords:", edit_kaomoji.keywords)
    kaomojitool.database.update_kaomoji(edit_kaomoji)

    print("Writing db", kaomojitool.database.filename)
    kaomojitool.database.write()


###############################################################################
# diff                                                                        #
###############################################################################
@cli.command()
@database_filename_option
@other_database_filename_option
@diff_type_option
@config_filename_option
def diff(database_filename, other_database_filename, diff_type,
         config_filename):
    """Compare two databases and logs the difference between them."""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    kaomojitool_other =\
        KaomojiTool(cli_database_filename=other_database_filename,
                    cli_config_filename=config_filename)

    if not kaomojitool.database or not kaomojitool_other.database:
        raise KaomojiToolNoDatabase

    diff_dict = kaomojitool.database.compare(other=kaomojitool_other.database,
                                             diff_type=diff_type)
    print(diff_dict)

    # TAKE SHA256, FORMAT FOR kaomojitool.database, WRITE


###############################################################################
# dbstatus                                                                    #
###############################################################################
@cli.command()
@database_filename_option
@config_filename_option
def dbstatus(database_filename, config_filename):
    """Show data from the database."""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    number_of_kaomoji = len(kaomojitool.database.kaomojis)

    print("database file:", kaomojitool.database.filename)
    print("number of kaomojis:", number_of_kaomoji)


###############################################################################
# query                                                                         #
###############################################################################
@cli.command()
@database_filename_option
@query_string_option
@config_filename_option
def query(database_filename, query_string, config_filename):
    """Queries the database for the keyword"""

    kaomojitool = KaomojiTool(cli_database_filename=database_filename,
                              cli_config_filename=config_filename)

    matches = kaomojitool.database.query(query_string)
    print(matches)

###############################################################################
# scrape                                                                      #
###############################################################################
@cli.command()
@url_to_scrape_option
@kaomoji_xpath_string_option
@keywords_xpath_string_option
@container_xpath_string_option
@config_filename_option
def scrape(url_to_scrape, kaomoji_xpath_string, keywords_xpath_string,
           container_xpath_string, config_filename):
    """Scrape kaomoji on websites via XPath.

    NON-IMPLEMENTED: Also, can scrape keywords and add them to kaomoji.

    Redirect the output to a file then add using 'kaomojitool add'.

    $ kaomojitool scrape -u <HTTPS://WEBSITE_ADDRESS> -x <XPATH> > MY_OUTPUT
    $ cat MY_OUTPUT | kaomojitool add

    First check output for broken/multiline items.
    """

    # https://devhints.io/xpath

    url = url_to_scrape
    page = requests.get(url)
    tree = html.fromstring(page.content)
    nds = tree.xpath(kaomoji_xpath_string)

    for kaomoji in nds:
        print(kaomoji.text_content())




if __name__ == "__main__":

    cli()



import base64
from hashlib import sha256
from typing import Union

# let's make a notebook of exceptions here: we can use it later or not

class KaomojiDBKaomojiExists(Exception):
    description="Kaomoji already exists in the KaomojiDB"
    def __init__(self, *args, **kwargs):
        super().__init__(self.description, *args, **kwargs)

class KaomojiDBKaomojiDoesntExist(Exception):
    description="Kaomoji doesn't exist in the KaomojiDB"
    def __init__(self, *args, **kwargs):
        super().__init__(self.description, *args, **kwargs)

class KaomojiKaomojiKeywordExists(Exception):
    description="The keyword already exists in the Kaomoji"
    def __init__(self, *args, **kwargs):
        super().__init__(self.description, *args, **kwargs)

class KaomojiKaomojiKeywordDoesntExist(Exception):
    description="The keyword doesn't already exists in the Kaomoji"
    def __init__(self, *args, **kwargs):
        super().__init__(self.description, *args, **kwargs)


class Kaomoji:
    """Represents a Kaomoji entity."""

    code: str = str()  # unicode of the kaomoji
    keywords: list[str] = list()  # list of strings

    def __init__(self, code=None, keywords=None, line_entry=None):

        if line_entry:
            self.from_line_entry(line_entry=line_entry)

        elif code:
            self.code = code
            self.add_keywords(keywords)

        self._make_inits()

    def add_keyword(self, keyword: str) -> None:
        """Adds a keyword to this kaomoji entity."""

        if keyword and not keyword in self.keywords:
            self.keywords.append(keyword.strip())

    def add_keywords(self, keywords: Union[str, list, None]= None) -> None:

        if not keywords:
            keyword_list = []
        elif isinstance(keywords, str):
            keyword_list = [kw.strip() for kw in keywords.split(',')]
        elif isinstance(keywords, list):
            keyword_list = keywords
        else:
            raise TypeError("keywords is not str | list")

        resume = list(set(self.keywords + keyword_list))  # remove duplicates

        while '' in resume:
            resume.remove('')

        self.keywords = resume

    def from_line_entry(self, line_entry: str):
        """Formats the database line entry as a Kaomoji instance."""

        self.code = str()  # unicode of the kaomoji
        self.keywords = list()  # list of strings

        line = line_entry.strip()

        code, *keywords_str = line.split('\t', maxsplit=1)
        self.code = code
        self.add_keywords(*keywords_str)

        self._make_inits()

        return self


    def remove_keyword(self, keyword: str) -> None:
        """Removes a keyword to this kaomoji entity."""

        if keyword in self.keywords:
            # self.keywords.remove(keyword.strip())
            self.keywords.remove(keyword.strip())

    def remove_keywords(self, keywords: Union[str, list, None]) -> None:

        if not keywords:
            keyword_list = []
        elif isinstance(keywords, str):
            keyword_list = [kw.strip() for kw in keywords.split(',')]
        elif isinstance(keywords, list):
            keyword_list = keywords
        else:
            raise TypeError("keywords is not str | list")

        for keyword in keyword_list:
            kw = keyword.strip()

            print("KEYWORD", kw)
            print("self.KWYWORDS", self.keywords)
            if kw in self.keywords:
                self.keywords.remove(kw)

    def to_line_entry(self, self_register=False) -> str:
        """Formats the current Kaomoji instance as a database line entry."""

        code = self.code
        keywords_str = ", ".join([keyword.strip()
                                  for keyword in self.keywords])
        line_entry = "{code}\t{keywords_str}\n"\
            .format(code=code, keywords_str=keywords_str)

        if self_register:
            self.line_entry = line_entry

        return line_entry

    def matches_query(self, query):
        for keyword in self.keywords:
            if keyword.startswith(query):
                return True

        return False


    def _make_hash(self, code, self_register=False) -> int:
        """Gives a UUID for a given kaomoji, for comparison.

        It is the base10 of the sha256 digest of the kaomoji code:
            HASH = INT10(SHA256(BYTES(UNICODE_KAOMOJI_CODE_UTF8)))

        With this we can know if some emoji is already on the DATABASE, so to
            append keywords to it.
        """

        code_bytes = code.encode("utf-8")
        the_hash = sha256(code_bytes).hexdigest()

        # the_hash = int(code_sha256_hex_digestion, base=16)

        if self_register:
            self.hash = the_hash

        return the_hash

    # def _make_hash(self, code, self_register=False) -> int:
    #     """Gives a UUID for a given kaomoji, for comparison.
    #
    #     It is the base10 of the sha256 digest of the kaomoji code:
    #         HASH = INT10(SHA256(BYTES(UNICODE_KAOMOJI_CODE_UTF8)))
    #
    #     With this we can know if some emoji is already on the DATABASE, so to
    #         append keywords to it.
    #     """
    #
    #     code_bytes = code.encode("utf-8")
    #     code_sha256_hex_digestion = sha256(code_bytes).hexdigest()
    #
    #     the_hash = int(code_sha256_hex_digestion, base=16)
    #
    #     if self_register:
    #         self.hash = the_hash
    #
    #     return the_hash


    def _make_shortcode(self, code, self_register=False) -> str:
        """Generates a base64 shortcode to the kaomoji's code
        """

        code_bytes = code.encode("utf-8")

        b64_shortcode = base64.b64encode(code_bytes)

        if self_register:
            self.shortcode = b64_shortcode

        return b64_shortcode

    def _make_inits(self):
        if self.code:
            self.hash = self._make_hash(code=self.code)
            self.shortcode = self._make_shortcode(code=self.code)

    def _hash_to_shortcode(self):
        pass

    def _shortcode_to_hash(self):
        pass

    def __eq__(self, other):
        """ Implements the == (equality) operator to compare two Kaomoji
                instances.
        """

        return hash(self) == hash(other)

    def __hash__(self):
        """ Implements hash() which makes a given emoji to be compared as a
                numeric entity.
        """

        return self.hash

    def __repr__(self):
        """ Implements a repr() pythonic programmatic representation of the
                Kaomoji entity.
        """

        return "<Kaomoji `{}'; kwords: `{}'>".format(self.code, self.keywords)

    def __str__(self):
        """Implements a str() string representation of the kaomoji."""

        return self.code


class KaomojiDB:
    """Offers facilities to edit and check the DB file."""

    def __init__(self, filename=None) -> None:
        """
        Args:
            filename (str): The filename of the splatmoji database to be read.

        Attributes:
            filename (str): the filename of the database file.
            kaomojis (dict): a dictionary which the key is the kaomoji code,
                and the value is a list of keywords, eg.:
                    kaomojis = dict({'o_o': ['keyword 1', 'keyword 2'],
                                    '~_o': ['keywordd', 'keyy2', 'etc']})
        """
        if filename:
            self.load_file(filename=filename)

    def load_file(self, filename: str) -> None:
        """ Loads a db file reading it in the format usable by KaomojiDB class.
        """

        self.filename = filename
        self.kaomojis = dict()
        self.entry_num = int()

        # TODO: check here for errors
        db_file = open(filename, "r")
        #with open(filename) as dbfile:
        lines = db_file.readlines()

        for line in lines:

            # processed_line = line.strip().split("\t", maxsplit=1)
            # if len(processed_line) == 2:
            #     code, keywords = processed_line
            # else:
            #     code, *throwaway = processed_line
            #     keywords = None
            #
            # kaomoji = Kaomoji(code=code, keywords=keywords)
            # kaomoji = Kaomoji().from_line_entry(line_entry=line)
            kaomoji = Kaomoji(line_entry=line)

            self.kaomojis.update({kaomoji.code: kaomoji})

        self.entry_num = len(self.kaomojis)

    def write(self, filename: str=None) -> None:
        """Writes a db file with the changes made."""

        if not filename:
            filename = self.filename

        database_file = open(filename, "w")

        for kaomoji in self.kaomojis.values():
            database_entry_line = kaomoji.to_line_entry()
            database_file.write(database_entry_line)

        database_file.close()
        self.load_file(filename=self.filename)

    def kaomoji_exists(self, kaomoji: Kaomoji) -> bool:
        """Checks if a kaomoji exists already in the database."""

        if kaomoji.code in self.kaomojis:
                return True

        return False

    def add_kaomoji(self, kaomoji: Kaomoji) -> Kaomoji:
        """Adds a Kaomoji to the database."""

        self.kaomojis.update({kaomoji.code: kaomoji})

        return self.kaomojis[kaomoji.code]

    def get_kaomoji(self, by_entity: Union[Kaomoji, str, int]) ->\
                                                        Union[Kaomoji, None]:

        # Kaomoji entity
        if isinstance(by_entity, Kaomoji):

            if by_entity.code in self.kaomojis:
                return self.kaomojis[by_entity.code]

        # str, unicode
        elif isinstance(by_entity, str):

            if by_entity in self.kaomojis:
                return self.kaomojis[by_entity]

        # int, numeric hash
        elif isinstance(by_entity, int):

            for kaomoji in self.kaomojis.values():
                if the_hash == kaomoji.hash:
                    return kaomoji
        else:
            raise TypeError("by_entity is not Kaomoji | str | int")

        return None

    def get_kaomoji_by_code(self, code: str) -> Union[Kaomoji, None]:
        """Gets a Kaomoji with it's current keywords from the database."""

        if code in self.kaomojis:
            return self.kaomojis[code]

        return None

    def get_kaomoji_by_kaomoji(self, other: Kaomoji) -> Union[Kaomoji, None]:
        """Gets a Kaomoji with it's current keywords from the database."""

        if other.code in self.kaomojis:
            return self.kaomojis[other.code]

        return None

    def get_kaomoji_by_hash(self, the_hash: int) -> Union[Kaomoji, None]:
        """Gets a Kaomoji with it's current keywords from the database."""

        for kaomoji in self.kaomojis.values():
            if the_hash == kaomoji.hash:
                return kaomoji

        return None

    def remove_kaomoji(self, kaomoji: Kaomoji) -> None:
        """Removes a Kaomoji from the database."""

        if kaomoji.code in self.kaomojis:
            self.kaomojis.pop(kaomoji.code)

    def update_kaomoji(self, kaomoji: Kaomoji) -> Kaomoji:
        """Updates keywords to database."""

        self.kaomojis.update({kaomoji.code: kaomoji})

        return self.kaomojis[kaomoji.code]

    def query(self, query):
        results = dict()

        for code, kaomoji in self.kaomojis.items():
            if kaomoji.matches_query(query):
                results.update({code: kaomoji})

        return results

    def compare(self, other, diff_type="additional") -> dict[str: Kaomoji]:
        """Compares two KaomojiDB instances."""

        if not isinstance(other, KaomojiDB):
            raise TypeError("other is not a KaomojiDB")

        other_extra_dict = dict()
        self_extra_dict = dict()
        difference_dict = dict()

        # other_extra - other has and self doesn't
        for kaomoji_code in other.kaomojis:
            if kaomoji_code not in self.kaomojis:
                other_extra_kaomoji = other.kaomojis[kaomoji_code]
                other_extra_dict.update({kaomoji_code: other_extra_kaomoji})

            else:
                for keyword in other.kaomojis[kaomoji_code].keywords:
                    if not keyword in self.kaomojis[kaomoji_code].keywords:
                        if not kaomoji_code in other_extra_dict:
                            other_extra_kaomoji = Kaomoji(code=kaomoji_code, keywords=list())
                            other_extra_dict.update({kaomoji_code: other_extra_kaomoji})
                        other_extra_dict[kaomoji_code].keywords.append(keyword)

        # self_extra - self has and other doesn't
        for kaomoji_code in self.kaomojis:
            if kaomoji_code not in other.kaomojis:
                self_extra_kaomoji = self.kaomojis[kaomoji_code]
                self_extra_dict.update({kaomoji_code: self_extra_kaomoji})

            else:
                for keyword in self.kaomojis[kaomoji_code].keywords:
                    if not keyword in other.kaomojis[kaomoji_code].keywords:
                        if not kaomoji_code in self_extra_dict:
                            self_extra_kaomoji = Kaomoji(code=kaomoji_code, keywords=list())
                            self_extra_dict.update({kaomoji_code: self_extra_kaomoji})
                        self_extra_dict[kaomoji_code].keywords.append(keyword)

        # difference between both
        difference_dict.update(other_extra_dict)
        difference_dict.update(self_extra_dict)

        diff_types = {
            "additional": other_extra_dict,
            "difference": difference_dict,
            "exclusive": self_extra_dict,
        }

        return diff_types[diff_type]
